"""this package is all the cross-UI "Items" for displaying
all of the things needed in an installer, a README, License, etc...
and a tool for actually installing itself.
"""

from enum import Enum
import getpass
import sys
import subprocess
import os


import colorama

sys.path.append('..')

try:
    from .baseitem import Item
    from .tree import Tree
except SystemError:
    from baseitem import Item
    from tree import Tree


class _CaseInsensitiveDict(dict):

    def __getitem__(self, key):
        return super(_CaseInsensitiveDict, self).__getitem__(key.lower())

    def __setitem__(self, key, val):
        super(_CaseInsensitiveDict, self).__setitem__(key.lower())


class Readme(Item):
    """
    displays a readme.
    """

    def __init__(self, file):
        self.file = file

    def run(self):
        for i in self.file:
            self.ui.echo(i)

        getpass.getpass('press enter to continue...')
        return 0


class Licence(Item):
    """
    displays a licence, and asks if it is OK
    """

    def __init__(self, file):
        self.file = file

    def run(self):
        self.ui.echo('LICENCE:', fcolour=colorama.Fore.RED)
        for i in self.file:
            self.ui.echo(i.strip())

        i = ''
        while i not in 'ynYN':
            i = self.ui.get_input('accept? (y/n): ')

        if i in 'yY':
            return 0

        return 1


class Caller(Item):
    """calls an external program"""

    def __init__(self, args):
        self.args = args

    def run(self):
        subprocess.call(self.args)
        return 0


class Choice(Item):
    """asks what they would like to do, and acts accordingly"""

    def __init__(self, msg="continue?", opts={'y': lambda: 0, 'n': lambda: 1}, ignorecase=True):
        self.msg = self.form(msg, opts)
        self.opts = _CaseInsensitiveDict(**opts) if ignorecase else opts

    @staticmethod
    def form(msg, opts):
        """forms the message to be sent"""
        return '{} ({}): '.format(msg, '/'.join(opts.keys()))

    def run(self):
        opt = ''
        while opt not in self.opts.keys():
            opt = self.ui.get_input(self.msg)

        return self.opts[opt]()


# class Installer(Item):
#     """actually installs the things."""

#     def __init__(self,
#         install_dir,
#         path_to_main=None,:
#         """
#         install_dir is the directory for the thing to be installed.

#         self.dir_tree = map(Tree, data_dirs)
#         self.path_to_main


if __name__ == '__main__':
    Choice().run()



"""main module for compiling scripts together, with all their data."""

import argparse
import sys
import os
import subprocess

from .archive_data import ArchiveHandler
from . import compiler


def main():
    """parse arguments and act accordingly"""

    parser = argparse.ArgumentParser(
        description="bundle installer scripts and their data",
        prog='tbip')
    parser.add_argument('script', help='script to compile with data')
    parser.add_argument('-d', '--data', dest='data',
                        help='data directory to bundle with script')
    parser.add_argument('-o', dest='output',
                        help='output filename', default='install.py')
    parser.add_argument(
        '-f', '--freeze',
        help='freeze the output script with pyinstaller',
        action='store_true')
    parser.add_argument('-u', '--upx-dir',
                        help='directory to upx.', dest='upx')

    parser.add_argument(
        '-w', '--windowed',
        help='windows specific: make no console window display.',
        action='store_true')

    parser.add_argument(
        '-i', '--icon', help='windows specific: use this icon in the executable')

    parser.add_argument('-v', '--version',
                        action='version', version='0.0 alpha')

    ns = parser.parse_args()

    print('creating, reading and formatting zip file, please wait...'
          '(this may take a while)')
    with open('example.so', 'w') as d:
        d.write(repr(ns))
    make(ns.script, ns.data, ns.output)

    if ns.freeze:
        commands = [
            os.path.join(os.path.dirname(sys.executable),
                         'Scripts', 'pyinstaller')
            if sys.platform == 'win32' else 'pyinstaller',
            ns.output,
            '--onefile',
            '--upx-dir=' + ns.upx if ns.upx is not None else '/',
        ]

        if sys.platform == 'win32':
            commands.append('-w' if ns.windowed else '-c')

        if sys.platform == 'win32' and ns.icon:
            commands.append('-i=' + ns.icon)

        print('running pyinstaller on %s with args %s' %
              (ns.script, ' '.join(commands)))
        subprocess.call(commands)


def make(file, data, dest):
    """combines data with file and puts it in dest."""
    try:
        openfile = open(file)

    except FileNotFoundError:
        print('%s is not found' % file, file=sys.stderr)
        sys.exit(1)

    except PermissionError:
        if os.path.isdir(file):
            print('%s is a directory!' % file, file=sys.stderr)
            sys.exit(1)

        print('Permission denied to %s' % file, file=sys.stderr)
        sys.exit(1)

    compiler.ScriptCompiler(openfile.read(), ArchiveHandler(data)).dump(dest)

main()



"""
compiler.py- this module attempts to bring a python script
along with it's data.
"""


__author__ = 'Michael Gill <michaelveenstra12@gmail.com>'
__version__ = '0.0a'

import base64
import io

try:
    from archive_data import ArchiveHandler

except ImportError:
    from .archive_data import ArchiveHandler

SCRIPT_SKELETON = '''

from zipfile import ZipFile
from tarfile import TarFile
import os
import base64
import colorama

if os.name == 'nt': colorama.init()


_PYTHON_INSTALL_TMP_DIR = os.environ['TMP'] if os.name == 'nt' else '/tmp/'
_PYTHON_INSTALL_ARCHIVE_FORMAT = '{format}'

_PYTHON_INSTALL_ARCHIVE_OBJECT = TarFile if _PYTHON_INSTALL_ARCHIVE_FORMAT == 'tar' else ZipFile

_PYTHON_INSTALL_DATA_BIN = #%here goes the data!%
print(colorama.Fore.CYAN, 'building archive...', colorama.Style.RESET_ALL)
with open(os.path.join(
  _PYTHON_INSTALL_TMP_DIR, '~$archive.zip'), 'wb') as f:
    f.write(base64.b64decode(_PYTHON_INSTALL_DATA_BIN.encode()))
print('done')

{lower_code}

'''


class ScriptCompiler:
    """compiles the script and a data directory into one file.
    will extract to {tmp}/~$archive.tmp on runtime.
    """

    def __init__(self, script, data: ArchiveHandler):
        self.script = script
        self.data = data

    def get_data(self,):
        """return text for an installer script"""
        dictionary = {
            'format': self.data.archive_type,
            'lower_code': self.script,
        }
        code = SCRIPT_SKELETON.format(**dictionary)
        upper, lower = code.split('#%here goes the data!%')
        raw = base64.b64encode(
            self.data.get_binary_data())
        with open('.tempfile', 'wb') as openfile:
            openfile.write(upper.encode())
            openfile.write(b"r'''")
            openfile.write(_break(raw))
            openfile.write(b"'''")
            openfile.write(lower.encode())

        code = open('.tempfile', 'rb').read()

        return code

    def dump(self, filename):
        """send the contents of this file to filename"""
        with open(filename, 'wb') as openfile:
            openfile.write(self.get_data())


# def _break(bytes_, num=50):
#     """breaks string at every numth character, and inserts a newline."""
#     pos = 0
#     new = b''
#     for char in bytes_:
#         if pos == num:
#             new += chr(char).encode() + b'\n'
#             pos = 0
#         new += chr(char).encode()
#         pos += 1

#     return new


def _break(bytes_, num=50):
    """breaks string at every numth character, and inserts a newline."""
    new = b''
    old = 0
    for i in range(0, len(bytes_), num):
        new += bytes_[old:i] + b'\n'
        old = i
    new += bytes_[old:]
    return new


if __name__ == '__main__':
    print(_break_test(b'1' * 500).decode())
    # ScriptCompiler(open('archive_data.py').read(),
    #                ArchiveHandler('test-dir')).dump('example.py')



"""
this file will create the installer itself.
"""

from enum import Enum
import shutil
import os
import sys
import zipfile

_MS_WINDOWS = sys.platform == 'win32'
TMP = os.environ['TMP'] if _MS_WINDOWS else '/tmp'

# error message to be shown if the archive does not exist in
# {tmp}/~$archive.tmp. this means that it has not been extracted from the
# script, as it would if the script was a proper tbip file, run through
# the compiler.
ERR_MSG = """the file {} does not exist.
this means that whoever wrote this script, (hopefully you) has not run your installer
script through the compiler yet, which bundles the data with the script. run
`tbip <installer script name> -d <data directory>` to bundle these together. if you are
not the developer, then contact the actual developer and tell them to fix it, or try
and fix it yourself :)""".format(os.path.join(TMP, '~$archive.zip'))


class ProgressUtils(Enum):
    """enumerations of methods of displaying progress."""
    PROGRESS_BAR = 0
    PERCENT = 1
    FILES = 2
    NULL = 3
    MSG = 4


class Installer:
    """script to create the installer itself."""

    def __init__(
            self,
            pre_install_items,
            post_install_items,
            install_dir,
            ui,
            install_progress=ProgressUtils.PROGRESS_BAR,
    ):
        """items is a sequence of all items that will
        be run in the installation process. ui is the user interface that
        all the items will be run in. application_name is the name of the application.
        install_progress is an enumeration value,
        in ProgressUtils. they are self-explanitory."""

        self.install_dir = install_dir
        self.post_install_items = post_install_items
        self.pre_install_items = pre_install_items
        self.ui = ui
        self.install_progress = install_progress
        for item in post_install_items + pre_install_items:
            item.set(ui)

    def run(self):
        """run all the items given in the proper UI, and
        install.
        """

        for item in self.pre_install_items:
            self.ui.process(item.run())

        self._extract()

        for item in self.post_install_items:
            self.ui.process(item.run())

    def _extract(self):
        """extract the items"""
        try:
            print(os.path.join(TMP, '~$archive.zip'))
            zipf = zipfile.ZipFile(os.path.join(TMP, '~$archive.zip'))

        except FileNotFoundError:
            print(ERR_MSG, file=sys.stderr)
            sys.exit(1)
        dest = self.install_dir
        meter = self.install_progress

        sendf = False

        if meter == ProgressUtils.PROGRESS_BAR:
            meter = self.ui.ProgressBar

        elif meter == ProgressUtils.PERCENT:
            meter = self.ui.Percent

        elif meter == ProgressUtils.FILES:
            sendf = True
            meter = lambda x: print('extracted', x)

        elif meter == ProgressUtils.MSG:
            print('extracting files...')
            meter = lambda _: None
            sendf = True

        elif meter == ProgressUtils.NULL:
            sendf = True
            meter = lambda _: None

        else:
            raise ValueError('invalid progress shower')

        # check that the directory exists
        if os.path.isdir(dest):
            shutil.rmtree(dest)

        # if the file exists, and is a file, remove it!
        if os.path.exists(dest):
            os.remove(dest)

        contents = zipf.infolist()

        # initialize meter with the amount of
        if not sendf:
            meter = meter(len(contents))

        for i, item in enumerate(contents):
            zipf.extract(item, dest)
            if sendf:
                meter(item.filename)
            else:
                meter.add(i)

        self.ui.clear()
